#include <msp/core/algorithm.h>
#include <msp/core/maputils.h>
#include <msp/strings/format.h>
#include "deviceinfo.h"
#include "error.h"
#include "sequence.h"
#include "sequencebuilder.h"
#include "sequencetemplate.h"

using namespace std;

namespace Msp {
namespace GL {

SequenceBuilder::SequenceBuilder(const SequenceTemplate &t):
	tmpl(t)
{
	const vector<SequenceTemplate::Step> &steps = tmpl.get_steps();
	for(vector<SequenceTemplate::Step>::const_iterator i=steps.begin(); i!=steps.end(); ++i)
		renderables[i->slot_name] = i->default_renderable;
	const vector<SequenceTemplate::PostProcessor> &postprocs = tmpl.get_postprocessors();
	for(SequenceTemplate::PostProcessorArray::const_iterator i=postprocs.begin(); i!=postprocs.end(); ++i)
		if(!i->slot_name.empty())
			postprocessors[i->slot_name] = 0;
}

void SequenceBuilder::set_renderable(const string &name, Renderable &rend)
{
	get_item(renderables, name) = &rend;
}

void SequenceBuilder::set_postprocessor(const string &name, PostProcessor &pproc)
{
	get_item(postprocessors, name) = &pproc;
}

void SequenceBuilder::set_debug_name(const string &name)
{
#ifdef DEBUG
	debug_name = name;
#else
	(void)name;
#endif
}

void SequenceBuilder::build(Sequence &sequence) const
{
#ifdef DEBUG
	if(!debug_name.empty())
		sequence.set_debug_name(debug_name);
#endif

	const vector<SequenceTemplate::Step> &steps = tmpl.get_steps();
	for(vector<SequenceTemplate::Step>::const_iterator i=steps.begin(); i!=steps.end(); ++i)
	{
		Renderable *renderable = get_item(renderables, i->slot_name);
		if(!renderable)
			continue;

		Sequence::Step &step = sequence.add_step(i->tag, *renderable);
		step.set_blend(i->blend);
		step.set_depth_test(i->depth_test);
		step.set_stencil_test(i->stencil_test);
		step.set_lighting(i->lighting);
	}

	const SequenceTemplate::PostProcessorArray &postprocs = tmpl.get_postprocessors();
	for(SequenceTemplate::PostProcessorArray::const_iterator i=postprocs.begin(); i!=postprocs.end(); ++i)
	{
		PostProcessor *proc = 0;
		if(!i->slot_name.empty())
			proc = get_item(postprocessors, i->slot_name);
		if(proc)
			sequence.add_postprocessor(*proc);
		else if(i->postprocessor_template)
		{
			proc = i->postprocessor_template->create(sequence.get_width(), sequence.get_height());
			if(proc)
			{
#ifdef DEBUG
				if(!debug_name.empty())
					proc->set_debug_name(format("%s/%d.pproc", debug_name, i-postprocs.begin()));
#endif
				sequence.add_postprocessor_owned(proc);
			}
		}
	}
}

Sequence *SequenceBuilder::build(unsigned w, unsigned h) const
{
	RefPtr<Sequence> sequence = new Sequence(w, h, create_frame_format());
	build(*sequence);
	return sequence.release();
}

Sequence *SequenceBuilder::build(const View &view) const
{
	RefPtr<Sequence> sequence = new Sequence(view, create_frame_format());
	build(*sequence);
	return sequence.release();
}

Sequence *SequenceBuilder::build(const Framebuffer &fbo) const
{
	RefPtr<Sequence> sequence = new Sequence(fbo, create_frame_format());
	build(*sequence);
	return sequence.release();
}

FrameFormat SequenceBuilder::create_frame_format() const
{
	unsigned samples = min(tmpl.get_maximum_multisample(), Limits::get_global().max_samples);
	if(samples<tmpl.get_required_multisample())
		throw invalid_operation("SequenceBuilder::create_frame_format");

	PixelComponents color_comp = (tmpl.get_alpha() ? RGBA : RGB);
	DataType color_type = (tmpl.get_hdr() ? HALF_FLOAT : UNSIGNED_BYTE);
	PixelFormat color_pf = make_pixelformat(color_comp, color_type);

	return (COLOR_ATTACHMENT,color_pf, DEPTH_ATTACHMENT).set_samples(samples);
}

} // namespace GL
} // namespace Msp
