#include <msp/core/maputils.h>
#include <msp/datafile/collection.h>
#include "ambientocclusion.h"
#include "blend.h"
#include "bloom.h"
#include "colorcurve.h"
#include "lighting.h"
#include "object.h"
#include "resources.h"
#include "scene.h"
#include "sequencetemplate.h"

using namespace std;

namespace Msp {
namespace GL {

SequenceTemplate::SequenceTemplate():
	hdr(false),
	alpha(false),
	required_multisample(0),
	max_multisample(0)
{ }

SequenceTemplate::~SequenceTemplate()
{
	for(PostProcessorArray::iterator i=postprocessors.begin(); i!=postprocessors.end(); ++i)
		delete i->postprocessor_template;
}


SequenceTemplate::PostProcessorRegistry &SequenceTemplate::get_postprocessor_registry()
{
	static PostProcessorRegistry registry;
	static bool initialized = false;
	if(!initialized)
	{
		registry.register_type<AmbientOcclusion>("ambient_occlusion");
		registry.register_type<Bloom>("bloom");
		registry.register_type<ColorCurve>("colorcurve");
		initialized = true;
	}
	return registry;
}


SequenceTemplate::Step::~Step()
{ }


SequenceTemplate::PostProcessor::PostProcessor(GL::PostProcessor::Template *ppt):
	postprocessor_template(ppt)
{ }


SequenceTemplate::PostProcLoader::PostProcLoader()
{
	get_postprocessor_registry().invoke_all(*this);
}


SequenceTemplate::Loader::Loader(SequenceTemplate &t, Collection &c):
	DataFile::CollectionObjectLoader<SequenceTemplate, Resources>(t, &c)
{
	add("hdr", &SequenceTemplate::hdr);
	add("alpha", &SequenceTemplate::alpha);
	add("multisample", &Loader::multisample);
	add("multisample", &Loader::multisample_range);
	add("postprocessor", &Loader::postprocessor);
	add("step", &Loader::step);
	add("step", &Loader::step_with_slot);

	// Deprecated
	add("pass", &Loader::step_with_slot);
}

void SequenceTemplate::Loader::postprocessor_loaded()
{
	obj.postprocessors.push_back(get_postprocessor_template());
}

void SequenceTemplate::Loader::multisample(unsigned samples)
{
	obj.required_multisample = samples;
	obj.max_multisample = samples;
}

void SequenceTemplate::Loader::multisample_range(unsigned req, unsigned max)
{
	obj.required_multisample = req;
	obj.max_multisample = max;
}

void SequenceTemplate::Loader::postprocessor(const string &slot)
{
	PostProcLoader ldr;
	load_sub_with(ldr);
	PostProcessor pp;
	pp.postprocessor_template = ldr.get_postprocessor_template();
	pp.slot_name = slot;
	obj.postprocessors.push_back(pp);
}

void SequenceTemplate::Loader::step(const string &tag)
{
	step_with_slot(tag, string());
}

void SequenceTemplate::Loader::step_with_slot(const string &tag, const string &rend)
{
	Step stp;
	stp.tag = tag;
	stp.slot_name = rend;
	if(coll)
	{
		Step::Loader ldr(stp, *coll);
		ldr.set_inline_base_name(format("%s/%d.step", get_source(), obj.steps.size()));
		load_sub_with(ldr);
	}
	else
		load_sub(stp);

	obj.steps.push_back(stp);
}


SequenceTemplate::Step::Loader::Loader(Step &p):
	DataFile::CollectionObjectLoader<Step>(p, 0)
{
	init();
}

SequenceTemplate::Step::Loader::Loader(Step &p, Collection &c):
	DataFile::CollectionObjectLoader<Step>(p, &c)
{
	init();
}

void SequenceTemplate::Step::Loader::init()
{
	add("blend", &Loader::blend);
	add("blend", &Loader::blend_factors);
	add("depth_test", &Loader::depth_test);
	add("depth_test", &Loader::depth_compare);
	add("lighting", &Loader::lighting);
	add("lighting", &Loader::lighting_inline);
	add("object", &Loader::object);
	add("scene", &Loader::scene);
	add("stencil_test", &Loader::stencil_test);
}

void SequenceTemplate::Step::Loader::set_inline_base_name(const string &n)
{
	inline_base_name = n;
}

void SequenceTemplate::Step::Loader::blend()
{
	load_sub(obj.blend);
}

void SequenceTemplate::Step::Loader::blend_factors(BlendFactor src, BlendFactor dest)
{
	obj.blend = Blend(src, dest);
}

void SequenceTemplate::Step::Loader::depth_test()
{
	load_sub(obj.depth_test);
}

void SequenceTemplate::Step::Loader::depth_compare(Predicate c)
{
	obj.depth_test = DepthTest(c);
}

void SequenceTemplate::Step::Loader::lighting_inline()
{
	RefPtr<Lighting> lightn = new Lighting;
	load_sub(*lightn);
	get_collection().add(inline_base_name+".lightn", lightn.get());
	obj.lighting = lightn.release();
}

void SequenceTemplate::Step::Loader::lighting(const string &name)
{
	obj.lighting = &get_collection().get<Lighting>(name);
}

void SequenceTemplate::Step::Loader::object(const string &name)
{
	obj.default_renderable = &get_collection().get<GL::Object>(name);
}

void SequenceTemplate::Step::Loader::scene(const string &name)
{
	obj.default_renderable = &get_collection().get<Scene>(name);
}

void SequenceTemplate::Step::Loader::stencil_test()
{
	load_sub(obj.stencil_test);
}

} // namespace GL
} // namespace Msp
